/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Written by Denis Lukianov [denis@voxelsoft.com] July 2003 - Royalty Free, No Copyright
/////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "CPEFileLoader.h"

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include <stdio.h>

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CPEFileLoader::Load(char *cFile)
{
	Unload();

	if(!cFile)
		return false;

	if(!strlen(cFile))
		return false;

	FILE *fp = fopen(cFile, "rb");

	if(!fp)
		return false;

	fseek(fp, OFFSET_OF_PE_HEADER_IN_STUB, SEEK_SET);

	if(!fread(&StubSize, sizeof(StubSize), 1, fp))
	{
		fclose(fp);
		Unload();
		return false;
	}

	fseek(fp, 0, SEEK_SET);

	pStub = new char[StubSize];

	if(!fread(pStub, StubSize, 1, fp))
	{
		fclose(fp);
		Unload();
		return false;
	}

	fseek(fp, StubSize, SEEK_SET);

	if(!fread(&PEHeader, sizeof(PEHeader), 1, fp))
	{
		fclose(fp);
		Unload();
		return false;
	}
	
	switch(PEHeader.Signature)
	{
	case IMAGE_NT_SIGNATURE:
		break;
	case IMAGE_DOS_SIGNATURE:
	case 0x0107454e:
	default:
		fclose(fp);
		Unload();
		return false;
	}

	if(PEHeader.SizeOfOptionalHeader)
	{
		unsigned int iLastPos = ftell(fp);

		if(PEHeader.SizeOfOptionalHeader > sizeof(sPEOptionalHeader))
		{
			fclose(fp);
			Unload();
			return false;
		}

		if(!fread(&PEOptionalHeader, min(PEHeader.SizeOfOptionalHeader, sizeof(sPEOptionalHeader)), 1, fp))
		{
			fclose(fp);
			Unload();
			return false;
		}

		fseek(fp, iLastPos + PEHeader.SizeOfOptionalHeader, SEEK_SET);
	}

	unsigned int iSections = PEHeader.NumberOfSections;
	pPESections = new sPESectionHeader[iSections];

	while(iSections--)
	{
		if(!fread(&pPESections[iSections], sizeof(sPESectionHeader), 1, fp))
		{
			fclose(fp);
			Unload();
			return false;
		}
	}

	iSections = PEHeader.NumberOfSections;
	pPESectionsData = new char *[iSections];
	memset(pPESectionsData, 0, sizeof(char *) * iSections);

	while(iSections--)
	{
		fseek(fp, pPESections[iSections].PointerToRawData, SEEK_SET);

		pPESectionsData[iSections] = new char[pPESections[iSections].SizeOfRawData];
		memset(pPESectionsData[iSections], 0, pPESections[iSections].SizeOfRawData);

		if(pPESections[iSections].SizeOfRawData)
		{
			if(!fread(pPESectionsData[iSections], pPESections[iSections].SizeOfRawData, 1, fp))
			{
				fclose(fp);
				Unload();
				return false;
			}
		}
	}

	fclose(fp);

	return true;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CPEFileLoader::LoadFromResource(HRSRC hResInfo)
{
	Unload();

	if(!hResInfo)
		return false;

	HGLOBAL hRes = LoadResource(GetModuleHandle(0), hResInfo);

	if(!hRes)
		return false;

	char *cSource = (char *)LockResource(hRes); 
	unsigned int iTotalSize = SizeofResource(GetModuleHandle(0), hResInfo);

	if(!cSource || !iTotalSize)
		return false;

	StubSize = *(unsigned int *)(cSource + OFFSET_OF_PE_HEADER_IN_STUB);

	pStub = new char[StubSize];

	memcpy(pStub, cSource, StubSize);
	memcpy(&PEHeader, cSource + StubSize, sizeof(PEHeader));

	switch(PEHeader.Signature)
	{
	case IMAGE_NT_SIGNATURE:
		break;
	case IMAGE_DOS_SIGNATURE:
	case 0x0107454e:
	default:
		Unload();
		return false;
	}

	char *cCurrentSourcePtr = cSource + StubSize + sizeof(PEHeader);

	if(PEHeader.SizeOfOptionalHeader)
	{
		if(PEHeader.SizeOfOptionalHeader > sizeof(sPEOptionalHeader))
		{
			Unload();
			return false;
		}

		unsigned int iActualOHeaderSize = (unsigned int)min(PEHeader.SizeOfOptionalHeader, sizeof(sPEOptionalHeader));
		memcpy(&PEOptionalHeader, cCurrentSourcePtr, iActualOHeaderSize);
		cCurrentSourcePtr += iActualOHeaderSize;
	}

	unsigned int iSections = PEHeader.NumberOfSections;
	pPESections = new sPESectionHeader[iSections];

	while(iSections--)
	{
		memcpy(&pPESections[iSections], cCurrentSourcePtr, sizeof(sPESectionHeader));
		cCurrentSourcePtr += sizeof(sPESectionHeader);
	}

	iSections = PEHeader.NumberOfSections;
	pPESectionsData = new char *[iSections];
	memset(pPESectionsData, 0, sizeof(char *) * iSections);

	while(iSections--)
	{
		cCurrentSourcePtr = cSource + pPESections[iSections].PointerToRawData;

		pPESectionsData[iSections] = new char[pPESections[iSections].SizeOfRawData];
		memset(pPESectionsData[iSections], 0, pPESections[iSections].SizeOfRawData);

		if(pPESections[iSections].SizeOfRawData)
		{
			memcpy(pPESectionsData[iSections], cCurrentSourcePtr,  pPESections[iSections].SizeOfRawData);
		}
	}

	return true;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CPEFileLoader::CalcOffsets()
{
	if(!pStub || !pPESections || !pPESectionsData)
		return false;

	unsigned int iOffset = 0;
	unsigned int iVirtualOffset = 0;

	iOffset += StubSize;
	iOffset += sizeof(sPEHeader);
	iOffset += (unsigned int)min(sizeof(sPEOptionalHeader), PEHeader.SizeOfOptionalHeader);

	unsigned int iSection = PEHeader.NumberOfSections;

	while(iSection--)
	{
		iOffset += sizeof(sPESectionHeader);
	}

	iSection = PEHeader.NumberOfSections;

	unsigned int iSizeOfImage = 0;
	unsigned int iSpecificSection = 0;

	iVirtualOffset = PEOptionalHeader.SizeOfHeaders;

	while(iSection--)
	{
		iOffset += iOffset % PEOptionalHeader.FileAlignment ? (PEOptionalHeader.FileAlignment - (iOffset % PEOptionalHeader.FileAlignment)) : 0;
		iVirtualOffset += iVirtualOffset % PEOptionalHeader.SectionAlignment ? (PEOptionalHeader.SectionAlignment - (iVirtualOffset % PEOptionalHeader.SectionAlignment)) : 0;

		pPESections[iSection].PointerToRawData = iOffset;
		pPESections[iSection].VirtualAddress = iVirtualOffset;
		
		iOffset += pPESections[iSection].SizeOfRawData;
		iVirtualOffset += pPESections[iSection].VirtualSize;

		unsigned int iVirtualSize = pPESections[iSection].VirtualSize / PEOptionalHeader.SectionAlignment;
		if(iVirtualSize * PEOptionalHeader.SectionAlignment < pPESections[iSection].VirtualSize)
			iVirtualSize++;
		
		iSizeOfImage += iVirtualSize * PEOptionalHeader.SectionAlignment;
	}

	iSizeOfImage += PEOptionalHeader.SizeOfHeaders;
	unsigned int iVirtualSizeOfImage = iSizeOfImage / PEOptionalHeader.SectionAlignment;
	if(iVirtualSizeOfImage * PEOptionalHeader.SectionAlignment < iSizeOfImage) /// @todo
		iVirtualSizeOfImage++;
	PEOptionalHeader.SizeOfImage = iVirtualSizeOfImage * PEOptionalHeader.SectionAlignment;

	return true;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CPEFileLoader::Save(char *cFile)
{
	if(!CalcOffsets())
		return false;

	if(!cFile)
		return false;

	if(!strlen(cFile))
		return false;

	FILE *fp = fopen(cFile, "wb");

	if(!fp)
		return false;

	if(!fwrite(pStub, StubSize, 1, fp))
	{
		fclose(fp);
		return false;
	}

	if(!fwrite(&PEHeader, sizeof(PEHeader), 1, fp))
	{
		fclose(fp);
		return false;
	}

	if(PEHeader.SizeOfOptionalHeader)
	{
		if(!fwrite(&PEOptionalHeader, min(sizeof(sPEOptionalHeader), PEHeader.SizeOfOptionalHeader), 1, fp))
		{
			fclose(fp);
			return false;
		}
	}

	unsigned int iSections = PEHeader.NumberOfSections;

	while(iSections--)
	{
		if(!fwrite(&pPESections[iSections], sizeof(sPESectionHeader), 1, fp))
		{
			fclose(fp);
			return false;
		}
	}

	iSections = PEHeader.NumberOfSections;
	unsigned int iOffset;

	char *cSpace = new char[PEOptionalHeader.FileAlignment];
	memset(cSpace, 0, PEOptionalHeader.FileAlignment);

	while(iSections--)
	{
		iOffset = ftell(fp);
		unsigned int Spacing = iOffset % PEOptionalHeader.FileAlignment ? (PEOptionalHeader.FileAlignment - (iOffset % PEOptionalHeader.FileAlignment)) : 0;

		if(Spacing)
		{
			if(!fwrite(cSpace, Spacing, 1, fp))
			{
				delete[] cSpace;
				fclose(fp);
				return false;
			}
		}

		if(pPESections[iSections].SizeOfRawData)
		{
			if(!fwrite(pPESectionsData[iSections], pPESections[iSections].SizeOfRawData, 1, fp))
			{
				delete[] cSpace;
				fclose(fp);
				return false;
			}
		}
	}

	delete[] cSpace;

	fclose(fp);

	return true;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CPEFileLoader::Unload()
{
	if(pStub)
	{
		delete[] pStub;
		pStub = 0;
	}

	if(pPESections)
	{
		delete[] pPESections;
		pPESections = 0;
	}

	/*if(pPEOldFixedSections)
	{
		delete[] pPEOldFixedSections;
		pPEOldFixedSections = 0;
	}*/

	if(pPESectionsData)
	{
		unsigned int iSections = PEHeader.NumberOfSections;
		while(iSections--)
		{
			if(pPESectionsData[iSections])
			{
				delete[] pPESectionsData[iSections];
				pPESectionsData[iSections] = 0;
			}
		}
		delete[] pPESectionsData;
		pPESectionsData = 0;
	}
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

CPEFileLoader::sPEHeader *CPEFileLoader::GetHeader()
{
	return &PEHeader;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

CPEFileLoader::sPEOptionalHeader *CPEFileLoader::GetOptionalHeader()
{
	return &PEOptionalHeader;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CPEFileLoader::GetSection(unsigned int iSection, sPESectionHeader &SectionHeader, char **SectionData)
{
	if(iSection >= PEHeader.NumberOfSections)
		return false;

	SectionHeader = pPESections[iSection];
	*SectionData = pPESectionsData[iSection];

	return true;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

/*bool CPEFileLoader::SetSection(unsigned int iSection, sPESectionHeader &NewSectionHeader, char *NewSectionData)
{
	if(iSection >= PEHeader.NumberOfSections)
		return false;

	pPESections[iSection] = NewSectionHeader;

	pPESections[iSection].SizeOfRawData = pPESections[iSection].VirtualSize;
	pPESections[iSection].SizeOfRawData += pPESections[iSection].SizeOfRawData % PEOptionalHeader.FileAlignment ? (PEOptionalHeader.FileAlignment - (pPESections[iSection].SizeOfRawData % PEOptionalHeader.FileAlignment)) : 0;

	unsigned int iSpace = pPESections[iSection].SizeOfRawData - pPESections[iSection].VirtualSize;

	if(pPESectionsData[iSection])
		delete[] pPESectionsData[iSection];

	pPESectionsData[iSection] = new char[pPESections[iSection].SizeOfRawData];
	memcpy(pPESectionsData[iSection], NewSectionData, pPESections[iSection].VirtualSize);
	memset(pPESectionsData[iSection] + pPESections[iSection].VirtualSize, 0, iSpace);

	// Re-calculate all sections and header
	CalcOffsets(pPEOldFixedSections, PEHeader.NumberOfSections);

	return true;
}*/

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

int CPEFileLoader::AppendNewSection(sPESectionHeader &NewSectionHeader, char *NewSectionData)
{
	int iSection = 0;
	PEHeader.NumberOfSections++;

	// Expand section headers
	sPESectionHeader *OldSections = pPESections;
	pPESections = new sPESectionHeader[PEHeader.NumberOfSections];
	memcpy(pPESections + 1, OldSections, sizeof(sPESectionHeader) * (PEHeader.NumberOfSections - 1));
	pPESections[iSection] = NewSectionHeader;
	delete[] OldSections;

	// Expand section data
	char **OldSectionsData = pPESectionsData;
	pPESectionsData = new char *[PEHeader.NumberOfSections];
	memcpy(pPESectionsData + 1, OldSectionsData, sizeof(char*) * (PEHeader.NumberOfSections - 1));
	delete[] OldSectionsData;

	iSection = 0;

	pPESections[iSection].SizeOfRawData = pPESections[iSection].VirtualSize;
	pPESections[iSection].SizeOfRawData += pPESections[iSection].SizeOfRawData % PEOptionalHeader.FileAlignment ? (PEOptionalHeader.FileAlignment - (pPESections[iSection].SizeOfRawData % PEOptionalHeader.FileAlignment)) : 0;

	unsigned int iSpace = pPESections[iSection].SizeOfRawData - pPESections[iSection].VirtualSize;

	pPESectionsData[iSection] = new char[pPESections[iSection].SizeOfRawData];
	memcpy(pPESectionsData[iSection], NewSectionData, pPESections[iSection].VirtualSize);
	memset(pPESectionsData[iSection] + pPESections[iSection].VirtualSize, 0, iSpace);

	// Re-calculate all sections and header
	CalcOffsets();

	return iSection;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CPEFileLoader::iGetCodeSection(unsigned int &iSection)
{
	return iGetCodeSectionUsingSectionList(iSection, pPESections, PEHeader.NumberOfSections);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CPEFileLoader::iGetSectionByVA(unsigned int iVA, unsigned int &iSection)
{
	return iGetSectionByVAUsingSectionList(iVA, iSection, pPESections, PEHeader.NumberOfSections);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CPEFileLoader::iGetCodeSectionUsingSectionList(unsigned int &iSection, sPESectionHeader *SectionList, unsigned int iSectionsInList)
{
	iSection = iSectionsInList;

	while(iSection--)
	{
		if(SectionList[iSection].Characteristics & SECTION_CHARACTERISTICS_MEM_READ
		&& SectionList[iSection].Characteristics & SECTION_CHARACTERISTICS_MEM_EXECUTE
		&& SectionList[iSection].Characteristics & SECTION_CHARACTERISTICS_CNT_CODE)
			return true; /// @todo should also check code section and entry point in the optional header, what if multiple code sections?
	}

	return false;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CPEFileLoader::iGetSectionByName(char *cName, unsigned int &iSection)
{
	iSection = PEHeader.NumberOfSections;

	while(iSection--)
	{
		if(strncmp(pPESections[iSection].ANSIName, cName, PE_SECTION_ANSI_NAME_LEN) == 0)
			return true;
	}

	return false;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CPEFileLoader::iGetSectionByVAUsingSectionList(unsigned int iVA, unsigned int &iSection, sPESectionHeader *SectionList, unsigned int iSectionsInList)
{
	iSection = iSectionsInList;

	while(iSection--)
	{
		if(SectionList[iSection].VirtualAddress <= iVA && SectionList[iSection].VirtualAddress + SectionList[iSection].SizeOfRawData > iVA)
			return true;
	}

	return false;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

CPEFileLoader::sPESectionHeader *CPEFileLoader::GetSectionHeader(unsigned int iSection)
{
	if(iSection >= PEHeader.NumberOfSections)
		return false;

	return &pPESections[iSection];
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

unsigned int CPEFileLoader::GetSectionAttributes(unsigned int iSection)
{
	if(iSection >= PEHeader.NumberOfSections)
		return false;

	return pPESections[iSection].Characteristics;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CPEFileLoader::SetSectionAttributes(unsigned int iSection, unsigned int iNewAttributes)
{
	if(iSection >= PEHeader.NumberOfSections)
		return false;

	pPESections[iSection].Characteristics = iNewAttributes;

	return true;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
